@@ -1,11 +1,13 @@
 import re, base64
-from django.template.context import BaseContext, RequestContext
+from django.conf import settings as django_settings
+from django.template.context import BaseContext, RequestContext, Context
 from django.template import (Parser, Lexer, Token,
     TOKEN_TEXT, COMMENT_TAG_START, COMMENT_TAG_END, TemplateSyntaxError)
 from django.utils.cache import cc_delim_re
 from django.utils.functional import Promise, LazyObject
 from django.http import HttpRequest
 from django.contrib.messages.storage.base import BaseStorage
+from django.utils.encoding import smart_str
 
 try:
     import cPickle as pickle
@@ -31,10 +33,43 @@ def second_pass_render(request, content):
             tokens = Lexer(bit, None).tokenize()
         else:
             tokens.append(Token(TOKEN_TEXT, bit))
-        context = RequestContext(request, unpickle_context(bit))
-        result.append(Parser(tokens).parse().render(context))
+
+        context = RequestContext(request,
+            restore_csrf_token(request, unpickle_context(bit)))
+        rendered = Parser(tokens).parse().render(context)
+
+        if settings.SECRET_DELIMITER in rendered:
+            rendered = second_pass_render(request, rendered)
+        result.append(rendered)
+
     return "".join(result)
 
+def restore_csrf_token(request, storage=None):
+    """
+    Given the request and a the context used during the second render phase,
+    this wil check if there is a CSRF cookie and restores if needed, to
+    counteract the way the CSRF framework invalidates the CSRF token after
+    each request/response cycle.
+    """
+    if storage is None:
+        storage = {}
+    try:
+        request.META["CSRF_COOKIE"] = request.COOKIES[django_settings.CSRF_COOKIE_NAME]
+    except KeyError:
+        csrf_token = storage.get('csrf_token', None)
+        if csrf_token:
+            request.META["CSRF_COOKIE"] = csrf_token
+    return storage
+
+def backup_csrf_token(context, storage=None):
+    """
+    Get the CSRF token and convert it to a string (since it's lazy)
+    """
+    if storage is None:
+        storage = Context()
+    storage['csrf_token'] = smart_str(context.get('csrf_token', 'NOTPROVIDED'))
+    return storage
+
 def drop_vary_headers(response, headers_to_drop):
     """
     Remove an item from the "Vary" header of an ``HttpResponse`` object.
@@ -66,12 +101,23 @@ def flatten_context(context, remove_lazy=True):
     e.g. lazy objects.
     """
     flat_context = {}
-    for context_dict in context.dicts:
-        if remove_lazy:
-            result = dict((k, v) for k, v in context_dict.iteritems() if not isinstance(v, forbidden_classes))
-            flat_context.update(result)
-        else:
-            flat_context.update(context_dict)
+    def _flatten(context):
+        if isinstance(context, dict):
+            for k, v in context.items():
+                if isinstance(context, BaseContext):
+                    _flatten(context)
+                else:
+                    flat_context[k] = v
+        elif isinstance(context, BaseContext):
+            for context_dict in context.dicts:
+                _flatten(context_dict)
+
+    # traverse the passed context and update the dictionary accordingly
+    _flatten(context)
+
+    if remove_lazy:
+        only_allowed = lambda dic: not isinstance(dic[1], forbidden_classes)
+        return dict(filter(only_allowed, flat_context.iteritems()))
     return flat_context
 
 def unpickle_context(content, pattern=None):
